# sage_setup: distribution = sagemath-combinat
import sage.dynamics.cellular_automata.catalog as cellular_automata

from sage.misc.lazy_import import lazy_import
lazy_import("sage.dynamics.cellular_automata.solitons",
            ["SolitonCellularAutomata", "PeriodicSolitonCellularAutomata"])
del lazy_import
