# sage_setup: distribution = sagemath-combinat
r"""
Partitions, tableaux, and others

Partitions
----------

- :ref:`sage.combinat.partition`
- :ref:`sage.combinat.skew_partition`
- :ref:`sage.combinat.partition_tuple`
- :ref:`sage.combinat.superpartition`
- :ref:`sage.combinat.tableau`
- :ref:`sage.combinat.tableau_tuple`
- :ref:`sage.combinat.skew_tableau`
- :ref:`sage.combinat.ribbon`
- :ref:`sage.combinat.ribbon_tableau`
- :ref:`sage.combinat.k_tableau`
- :ref:`sage.combinat.shifted_primed_tableau`

- :ref:`sage.combinat.tableau_residues`

RSK
---

- :ref:`sage.combinat.rsk`
- :ref:`sage.combinat.growth`
"""
