# sage_setup: distribution = sagemath-combinat
# from symmetrica import *

from sage.libs.symmetrica.symmetrica import start

# kostka
from sage.libs.symmetrica.symmetrica import kostka_number_symmetrica as kostka_number
from sage.libs.symmetrica.symmetrica import kostka_tab_symmetrica as kostka_tab
from sage.libs.symmetrica.symmetrica import kostka_tafel_symmetrica as kostka_tafel


# sab
from sage.libs.symmetrica.symmetrica import dimension_symmetrization_symmetrica as dimension_symmetrization
from sage.libs.symmetrica.symmetrica import bdg_symmetrica as bdg
from sage.libs.symmetrica.symmetrica import sdg_symmetrica as sdg
from sage.libs.symmetrica.symmetrica import odg_symmetrica as odg
from sage.libs.symmetrica.symmetrica import specht_dg_symmetrica as specht_dg
from sage.libs.symmetrica.symmetrica import ndg_symmetrica as ndg
# from symmetrica import glmndg_symmetrica as glmndg


# sc
from sage.libs.symmetrica.symmetrica import chartafel_symmetrica as chartafel
from sage.libs.symmetrica.symmetrica import charvalue_symmetrica as charvalue
from sage.libs.symmetrica.symmetrica import kranztafel_symmetrica as kranztafel
# from symmetrica import c_ijk_sn_symmetrica  as c_ijk_sn

# part
from sage.libs.symmetrica.symmetrica import strict_to_odd_part_symmetrica as strict_to_odd_part
from sage.libs.symmetrica.symmetrica import odd_to_strict_part_symmetrica as odd_to_strict
from sage.libs.symmetrica.symmetrica import q_core_symmetrica as q_core
from sage.libs.symmetrica.symmetrica import gupta_nm_symmetrica as gupta_nm
from sage.libs.symmetrica.symmetrica import gupta_tafel_symmetrica as gupta_tafel
from sage.libs.symmetrica.symmetrica import random_partition_symmetrica as random_partition


# schur
from sage.libs.symmetrica.symmetrica import outerproduct_schur_symmetrica as outerproduct_schur
from sage.libs.symmetrica.symmetrica import dimension_schur_symmetrica as dimension_schur
from sage.libs.symmetrica.symmetrica import part_part_skewschur_symmetrica as part_part_skewschur
from sage.libs.symmetrica.symmetrica import newtrans_symmetrica as newtrans
from sage.libs.symmetrica.symmetrica import compute_schur_with_alphabet_symmetrica as compute_schur_with_alphabet
from sage.libs.symmetrica.symmetrica import compute_homsym_with_alphabet_symmetrica as compute_homsym_with_alphabet
from sage.libs.symmetrica.symmetrica import compute_elmsym_with_alphabet_symmetrica as compute_elmsym_with_alphabet
from sage.libs.symmetrica.symmetrica import compute_monomial_with_alphabet_symmetrica as compute_monomial_with_alphabet
from sage.libs.symmetrica.symmetrica import compute_powsym_with_alphabet_symmetrica as compute_powsym_with_alphabet
from sage.libs.symmetrica.symmetrica import compute_schur_with_alphabet_det_symmetrica as compute_schur_with_alphabet_det

from sage.libs.symmetrica.symmetrica import t_SCHUR_MONOMIAL_symmetrica as t_SCHUR_MONOMIAL
from sage.libs.symmetrica.symmetrica import t_SCHUR_HOMSYM_symmetrica as t_SCHUR_HOMSYM
from sage.libs.symmetrica.symmetrica import t_SCHUR_POWSYM_symmetrica as t_SCHUR_POWSYM
from sage.libs.symmetrica.symmetrica import t_SCHUR_ELMSYM_symmetrica as t_SCHUR_ELMSYM

from sage.libs.symmetrica.symmetrica import t_MONOMIAL_SCHUR_symmetrica as t_MONOMIAL_SCHUR
from sage.libs.symmetrica.symmetrica import t_MONOMIAL_HOMSYM_symmetrica as t_MONOMIAL_HOMSYM
from sage.libs.symmetrica.symmetrica import t_MONOMIAL_POWSYM_symmetrica as t_MONOMIAL_POWSYM
from sage.libs.symmetrica.symmetrica import t_MONOMIAL_ELMSYM_symmetrica as t_MONOMIAL_ELMSYM

from sage.libs.symmetrica.symmetrica import t_ELMSYM_SCHUR_symmetrica as t_ELMSYM_SCHUR
from sage.libs.symmetrica.symmetrica import t_ELMSYM_MONOMIAL_symmetrica as t_ELMSYM_MONOMIAL
from sage.libs.symmetrica.symmetrica import t_ELMSYM_HOMSYM_symmetrica as t_ELMSYM_HOMSYM
from sage.libs.symmetrica.symmetrica import t_ELMSYM_POWSYM_symmetrica as t_ELMSYM_POWSYM

from sage.libs.symmetrica.symmetrica import t_HOMSYM_SCHUR_symmetrica as t_HOMSYM_SCHUR
from sage.libs.symmetrica.symmetrica import t_HOMSYM_MONOMIAL_symmetrica as t_HOMSYM_MONOMIAL
from sage.libs.symmetrica.symmetrica import t_HOMSYM_POWSYM_symmetrica as t_HOMSYM_POWSYM
from sage.libs.symmetrica.symmetrica import t_HOMSYM_ELMSYM_symmetrica as t_HOMSYM_ELMSYM

from sage.libs.symmetrica.symmetrica import t_POWSYM_SCHUR_symmetrica as t_POWSYM_SCHUR
from sage.libs.symmetrica.symmetrica import t_POWSYM_HOMSYM_symmetrica as t_POWSYM_HOMSYM
from sage.libs.symmetrica.symmetrica import t_POWSYM_ELMSYM_symmetrica as t_POWSYM_ELMSYM
from sage.libs.symmetrica.symmetrica import t_POWSYM_MONOMIAL_symmetrica as t_POWSYM_MONOMIAL


from sage.libs.symmetrica.symmetrica import mult_schur_schur_symmetrica as mult_schur_schur
from sage.libs.symmetrica.symmetrica import mult_monomial_monomial_symmetrica as mult_monomial_monomial


from sage.libs.symmetrica.symmetrica import hall_littlewood_symmetrica as hall_littlewood

from sage.libs.symmetrica.symmetrica import t_POLYNOM_POWER_symmetrica as t_POLYNOM_POWER
from sage.libs.symmetrica.symmetrica import t_POLYNOM_SCHUR_symmetrica as t_POLYNOM_SCHUR
from sage.libs.symmetrica.symmetrica import t_POLYNOM_ELMSYM_symmetrica as t_POLYNOM_ELMSYM
from sage.libs.symmetrica.symmetrica import t_POLYNOM_MONOMIAL_symmetrica as t_POLYNOM_MONOMIAL

from sage.libs.symmetrica.symmetrica import scalarproduct_schur_symmetrica as scalarproduct_schur

# plet
from sage.libs.symmetrica.symmetrica import plethysm_symmetrica as plethysm
from sage.libs.symmetrica.symmetrica import schur_schur_plet_symmetrica as schur_schur_plet

# sb
from sage.libs.symmetrica.symmetrica import mult_schubert_schubert_symmetrica as mult_schubert_schubert
from sage.libs.symmetrica.symmetrica import t_SCHUBERT_POLYNOM_symmetrica as t_SCHUBERT_POLYNOM
from sage.libs.symmetrica.symmetrica import t_POLYNOM_SCHUBERT_symmetrica as t_POLYNOM_SCHUBERT
from sage.libs.symmetrica.symmetrica import mult_schubert_variable_symmetrica as mult_schubert_variable
from sage.libs.symmetrica.symmetrica import divdiff_perm_schubert_symmetrica as divdiff_perm_schubert
from sage.libs.symmetrica.symmetrica import scalarproduct_schubert_symmetrica as scalarproduct_schubert
from sage.libs.symmetrica.symmetrica import divdiff_schubert_symmetrica as divdiff_schubert

start()
