# sage_setup: distribution = sagemath-combinat
cpdef tuple compute_new_lusztig_datum(list enhanced_braid_chain, initial_lusztig_datum)
cpdef tuple tropical_plucker_relation(tuple a, lusztig_datum)
cpdef list enhance_braid_move_chain(braid_move_chain, cartan_type)
