# sage_setup: distribution = sagemath-combinat
r"""
Catalog of crystal models For `B(\infty)`

We currently have the following models:

* :class:`AlcovePaths
  <sage.combinat.crystals.alcove_path.InfinityCrystalOfAlcovePaths>`
* :class:`GeneralizedYoungWalls
  <sage.combinat.crystals.generalized_young_walls.InfinityCrystalOfGeneralizedYoungWalls>`
* :class:`LSPaths <sage.combinat.crystals.littelmann_path.InfinityCrystalOfLSPaths>`
* :class:`Multisegments <sage.combinat.crystals.multisegments.InfinityCrystalOfMultisegments>`
* :class:`MVPolytopes <sage.combinat.crystals.mv_polytopes.MVPolytopes>`
* :class:`NakajimaMonomials <sage.combinat.crystals.monomial_crystals.InfinityCrystalOfNakajimaMonomials>`
* :class:`PBW <sage.combinat.crystals.pbw_crystal.PBWCrystal>`
* :class:`PolyhedralRealization <sage.combinat.crystals.polyhedral_realization.InfinityCrystalAsPolyhedralRealization>`
* :class:`RiggedConfigurations
  <sage.combinat.rigged_configurations.rc_infinity.InfinityCrystalOfRiggedConfigurations>`
* :class:`Star <sage.combinat.crystals.star_crystal.StarCrystal>`
* :class:`Tableaux <sage.combinat.crystals.infinity_crystals.InfinityCrystalOfTableaux>`
"""

from .generalized_young_walls import InfinityCrystalOfGeneralizedYoungWalls as GeneralizedYoungWalls
from .multisegments import InfinityCrystalOfMultisegments as Multisegments
from .monomial_crystals import InfinityCrystalOfNakajimaMonomials as NakajimaMonomials
from sage.combinat.rigged_configurations.rc_infinity import InfinityCrystalOfRiggedConfigurations as RiggedConfigurations
from .infinity_crystals import InfinityCrystalOfTableaux as Tableaux
from sage.combinat.crystals.polyhedral_realization import InfinityCrystalAsPolyhedralRealization as PolyhedralRealization
from sage.combinat.crystals.pbw_crystal import PBWCrystal as PBW
from sage.combinat.crystals.mv_polytopes import MVPolytopes
from sage.combinat.crystals.star_crystal import StarCrystal as Star
from sage.combinat.crystals.littelmann_path import InfinityCrystalOfLSPaths as LSPaths
from sage.combinat.crystals.alcove_path import InfinityCrystalOfAlcovePaths as AlcovePaths
