# sage_setup: distribution = sagemath-combinat

from sage.monoids.free_monoid import FreeMonoid
from sage.monoids.string_monoid import (BinaryStrings, OctalStrings, HexadecimalStrings,
                                        Radix64Strings, AlphabeticStrings)

from sage.monoids.free_abelian_monoid import FreeAbelianMonoid

from sage.monoids.string_ops import (
    strip_encoding,
    frequency_distribution,
    coincidence_index,
    coincidence_discriminant)
