# sage_setup: distribution = sagemath-combinat
r"""
Algebraic combinatorics

Thematic tutorials
------------------

- `Algebraic Combinatorics in Sage <../../../../thematic_tutorials/algebraic_combinatorics.html>`_
- `Lie Methods and Related Combinatorics in Sage <../../../../thematic_tutorials/lie.html>`_
- `Linear Programming (Mixed Integer) <../../../../thematic_tutorials/linear_programming.html>`_

Enumerated sets of combinatorial objects
----------------------------------------

- :ref:`sage.combinat.catalog_partitions`
- :class:`~sage.combinat.gelfand_tsetlin_patterns.GelfandTsetlinPattern`,
  :class:`~sage.combinat.gelfand_tsetlin_patterns.GelfandTsetlinPatterns`
- :class:`~sage.combinat.knutson_tao_puzzles.KnutsonTaoPuzzleSolver`

Groups and Algebras
-------------------

- :ref:`Catalog of algebras <sage.algebras.catalog>`
- :ref:`Groups <sage.groups.groups_catalog>`
- :class:`SymmetricGroup`, :class:`CoxeterGroup`, :class:`WeylGroup`
- :class:`~sage.combinat.diagram_algebras.PartitionAlgebra`
- :class:`~sage.algebras.iwahori_hecke_algebra.IwahoriHeckeAlgebra`
- :class:`~sage.combinat.symmetric_group_algebra.SymmetricGroupAlgebra`
- :class:`~sage.algebras.nil_coxeter_algebra.NilCoxeterAlgebra`
- :class:`~sage.algebras.affine_nil_temperley_lieb.AffineNilTemperleyLiebTypeA`
- :ref:`sage.combinat.descent_algebra`
- :ref:`sage.combinat.diagram_algebras`
- :ref:`sage.combinat.blob_algebra`

Combinatorial Representation Theory
-----------------------------------

- :ref:`sage.combinat.root_system.all`
- :ref:`sage.combinat.crystals.all`
- :ref:`sage.combinat.rigged_configurations.all`
- :ref:`sage.combinat.cluster_algebra_quiver.all`
- :class:`~sage.combinat.kazhdan_lusztig.KazhdanLusztigPolynomial`
- :class:`~sage.combinat.symmetric_group_representations.SymmetricGroupRepresentation`
- :ref:`sage.combinat.specht_module`
- :ref:`sage.combinat.yang_baxter_graph`
- :ref:`sage.combinat.hall_polynomial`
- :ref:`sage.combinat.key_polynomial`

Operads and their algebras
--------------------------

- :ref:`sage.combinat.free_dendriform_algebra`
- :ref:`sage.combinat.free_prelie_algebra`
- :ref:`sage.algebras.free_zinbiel_algebra`
"""
