# sage_setup: distribution = sagemath-combinat
r"""
Path tableaux

- :ref:`sage.combinat.path_tableaux.path_tableau`
- :ref:`sage.combinat.path_tableaux.dyck_path`
- :ref:`sage.combinat.path_tableaux.frieze`
- :ref:`sage.combinat.path_tableaux.semistandard`
"""
# install the docstring of this module to the containing package
from sage.misc.namespace_package import install_doc
install_doc(__package__, __doc__)
del install_doc
